package ru.dbotthepony.kstarbound.defs.projectile

import com.google.common.collect.ImmutableMap
import org.apache.logging.log4j.LogManager
import ru.dbotthepony.kstarbound.Starbound
import ru.dbotthepony.kstarbound.defs.AssembledPrototype
import ru.dbotthepony.kstarbound.defs.DamageType
import ru.dbotthepony.kstarbound.defs.FrameSet
import ru.dbotthepony.kstarbound.world.entities.projectile.AbstractProjectileMovementController
import ru.dbotthepony.kstarbound.world.entities.projectile.Projectile
import ru.dbotthepony.kvector.vector.Color

class ConfiguredProjectile(
	json: ImmutableMap<String, Any>,
	val projectileName: String,
	val physics: ProjectilePhysics,
	val damageKindImage: String?,
	val damageType: DamageType,
	val damageKind: String?,
	val pointLight: Boolean,
	val lightColor: Color?,
	val onlyHitTerrain: Boolean,
	val orientationLocked: Boolean,
	val image: FrameSet,
	val timeToLive: Double,
	val animationCycle: Double,
	val bounces: Int,
	val frameNumber: Int,
	val scripts: Array<String>,
	val actionOnReap: List<IActionOnReap>,
	val animationLoops: Boolean,
	val hydrophobic: Boolean,
	val piercing: Boolean,
	val speed: Double,
	val power: Double,
) : AssembledPrototype<ConfiguredProjectile, ConfigurableProjectile>(json) {
	override fun disassemble(): ConfigurableProjectile {
		TODO("Not yet implemented")
	}

	override fun toString(): String {
		return "ConfiguredProjectile($projectileName)"
	}
}

interface IActionOnReap {
	val name: String
	fun execute(projectile: Projectile)
}

data class CActionConfig(
	val file: String,
	val delegate: IActionOnReap?,
) : IActionOnReap {
	override val name: String = "config"

	override fun execute(projectile: Projectile) {
		delegate?.execute(projectile)
	}
}

data class CActionProjectile(
	val type: String,
	val angle: Double,
	val inheritDamageFactor: Double,
) : IActionOnReap {
	override val name: String = "projectile"

	override fun execute(projectile: Projectile) {
		val def = Starbound.projectilesAccess[type]

		if (def == null) {
			LOGGER.error("Tried to create unknown projectile '{}' as result of reap of '{}'!", type, projectile.def.projectileName)
			return
		}

		val ent = Projectile(projectile.world, def)
		ent.position = projectile.position
		// ent.angle = projectile.angle
		ent.angle = Math.toRadians(angle)

		if (ent.movement is AbstractProjectileMovementController) {
			ent.movement.push()
		}

		ent.spawn()
	}

	companion object {
		private val LOGGER = LogManager.getLogger(CActionProjectile::class.java)
	}
}

data class CActionSound(
	val options: List<String>
) : IActionOnReap {
	override val name: String = "sound"

	override fun execute(projectile: Projectile) {
		println("Play sound ${options.random()}!")
	}
}

data class CActionLoop(
	val count: Int,
	val body: List<IActionOnReap>
) : IActionOnReap {
	override val name: String = "loop"

	override fun execute(projectile: Projectile) {
		for (i in 0 until count) {
			for (action in body) {
				action.execute(projectile)
			}
		}
	}
}

data class CActionActions(
	val list: List<IActionOnReap>
) : IActionOnReap {
	override val name: String = "actions"

	override fun execute(projectile: Projectile) {
		for (action in list) {
			action.execute(projectile)
		}
	}
}
